#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'

import asyncio
import json

from aiopyramid.websocket.config.gunicorn import WebsocketMapper
from aiopyramid.websocket.view import WebsocketConnectionView
from pyramid.view import view_config

from pyams_notify_ws import _


users_lock = asyncio.Lock()
users = {}


class WebsocketSubscription(object):
    """Websocket user subscription

    A subscription is defined by:
    - a principal identified by it's ID, title, authentication token and principals
    - a set of contexts; each context is defined by a unique key ('*' for a global
    subscription) for which we subscribe for a list of events types
    """

    def __init__(self, principal):
        self.principal = principal
        self.contexts = {}

    def update_contexts(self, new_contexts):
        contexts = self.contexts
        for ctx, events in new_contexts.items():
            contexts[ctx] = set(contexts.get(ctx, set())) | set(events)
        self.contexts = contexts

    def filter_target(self, target):
        if not target:
            return True
        return self.principal['id'] in target


@view_config(route_name='subscribe', mapper=WebsocketMapper)
class WebsocketSubscribeView(WebsocketConnectionView):
    """Websocket subscribe view"""

    async def on_open(self):
        pass

    async def on_message(self, message):
        try:
            message = json.loads(message)
        except ValueError:
            pass
        else:
            action = message.get('action')
            if action == 'subscribe':
                principal = message.get('principal')
                translate = self.request.localizer.translate
                if not principal:
                    await self.ws.send(json.dumps({'status': 'error',
                                                   'error': translate(_("API Error")),
                                                   'message': translate(_("Missing principal argument!"))}))
                    return
                if self.request.unauthenticated_userid != principal.get('id'):
                    await self.ws.send(json.dumps({'status': 'error',
                                                   'error': translate(_("Authentication Error")),
                                                   'message': translate(_("Can't subscribe to notifications!"))}))
                    return
                async with users_lock:
                    subscription = users.get(self.ws)
                    if subscription is None:
                        subscription = WebsocketSubscription(principal)
                    subscription.update_contexts(message.get('context'))
                    users[self.ws] = subscription

    async def on_close(self):
        if self.ws in users:
            async with users_lock:
                del users[self.ws]
